import asyncio
import functools
import ourpipes.framework as ourpipes
from utils.async_functools import async_partial

def test_can_run_default_schema_executor_compressed():

    async def f(x,y):
        return x+y

    memory = {
        "a": 1,
        "b": 2,
        "c": 3
    }
    schema = {
        "x": ("f", (("a"), ("b", "y"))),
        "y": ("f", (("x", "x"), ("c", "y"))),
    }

    functions = {
        "f": f,
    }

    compression_name = "gzip_pickle"

    try:
        result = asyncio.run(
            ourpipes.execute_compressed_pipeline(
                functions, 
                compression_name, 
                ourpipes.compress_pipeline(
                    compression_name, 
                    schema, 
                    memory,
                    "y"
                ),
            ),
        )
        assert result == 6
    except Exception as e:
        assert False, str(e)

def test_can_run_default_schema_executor():

    async def f(x,y):
        return x+y

    memory = {
        "a": 1,
        "b": 2,
        "c": 3
    }
    schema = {
        "x": ("f", (("a", "x"), ("b", "y"))),
        "y": ("f", (("x", "x"), ("c", "y"))),
    }
    functions = {
        "f": f,
    }

    try:
        result = asyncio.run(ourpipes.execute_schema(functions, schema, memory, "y"))
        assert result == 6
    except Exception as e:
        assert False, str(e)

def test_run_default_schema_executor_should_fail():

    async def f(x,y):
        raise Exception("cannot compute, sorry...")

    memory = {
        "a": 1,
        "b": 2,
        "c": 3
    }
    schema = {
        "x": ("f", (("a", "x"), ("b", "y"))),
        "y": ("f", (("x", "x"), ("c", "y"))),
    }
    functions = {
        "f": f,
    }

    try:
        result = asyncio.run(ourpipes.execute_schema(functions, schema, memory, "y"))
        assert isinstance(result, Exception)
    except Exception as e:
        assert False, str(e)

def test_custom_function_executor():

    async def custom_executor(*fns, **kwargs):
        return [fn(*ars, **kws) for fn, ars, kws in fns]

    def custom_wrapper(fn_name: str, functions: dict, args, kwargs):
        return functions[fn_name], args, kwargs

    def f(x,y):
        return x+y

    memory = {
        "a": 1,
        "b": 2,
        "c": 3
    }
    schema = {
        "x": ("f", (("a", "x"), ("b", "y"))),
        "y": ("f", (("x", "x"), ("c", "y"))),
    }
    functions = {
        "f": f,
    }

    result = asyncio.run(ourpipes.execute_schema(functions, schema, memory, "y", paralell_function_executor=custom_executor, sequential_function_wrapper=custom_wrapper))
    assert result == 6

def test_run_schema_with_key_error_should_fail():

    async def f(x,y):
        return x+y

    memory = {
        "a": 1,
        "b": 2,
        "c": 3
    }
    schema = {
        "x": ("f", (("a", "y"), ("b", "y"))),
    }
    functions = {
        "f": f,
    }

    try:
        result = asyncio.run(ourpipes.execute_schema(functions, schema, memory))
    except Exception as e:
        assert True, str(e)    

def test_run_schema_executor_in_schema():

    async def f(x, y):
        return x+y

    async def g(x, y):
        return {x: y}

    async def h(x, y):
        return {**x, **y}

    async def extract(key: str, from_dict: dict):
        return from_dict[key]

    atomic_functions = {
        "f": f,
        "g": g,
        "h": h,
        "extract": extract,
    }
    async def wrap(schema, memory, key):
        return await ourpipes.execute_schema(atomic_functions, schema, memory, key)

    functions = {
        **atomic_functions,
        **{
            "e": wrap
        }
    }

    memory = {
        "a": 1,
        "b": 2,
        "c": {
            "a": 1
        },
        "d": {
            "f": ("f", (("a", "x"), ("x", "y"))),
        },
        "e": "x",
        "key": "f"
    }
    schema = {
        "x": ("f", (("a", "x"), ("b", "y"))),
        
        "g": ("g", (("e", "x"), ("x", "y"))),
        "h": ("h", (("c", "x"), ("g", "y"))),
        
        "y": ("e", (("d", "schema"), ("h", "memory"), ("key", "key"))),
    }

    result = asyncio.run(ourpipes.execute_schema(functions, schema, memory, "y"))
    assert result == 4

def test_run_schema_as_far_as_possible():

    async def f(x,y):
        return x+y

    memory = {
        "a": 1,
        "b": 2
    }
    schema = {
        "x": ("f", (("a", "x"), ("b", "y"))),
        "y": ("g", (("x", "x"),)),
    }
    functions = {
        "f": f,
    }

    try:
        executed = asyncio.run(ourpipes.execute_schema(functions, schema, memory))
        new_schema, new_memory = ourpipes.fast_forward_schema(schema, executed)
        assert "y" in new_schema and "x" not in new_schema
        assert "x" in new_memory and "a" not in new_memory and not "b" in new_memory

        # compute new scheme and memory when now g is available
        async def g(x):
            return x +1

        new_executed = asyncio.run(
            ourpipes.execute_schema({"g": g}, new_schema, new_memory)
        )
        new_new_schema, new_new_memory = ourpipes.fast_forward_schema(new_schema, new_memory)
        assert new_executed['y'] == 4
        assert new_new_schema == {}
        assert new_new_memory == {}
    except Exception as e:
        assert False, str(e)    

def test_run_schema_as_far_as_possible_while_can_be_completed():

    async def f(x,y):
        return x+y

    async def g(x):
        return x +1

    memory = {
        "a": 1,
        "b": 2
    }
    schema = {
        "x": ("f", (("a", "x"), ("b", "y"))),
        "y": ("g", (("x", "x"),)),
    }
    functions = {
        "f": f,
        "g": g,
    }

    try:
        executed = asyncio.run(ourpipes.execute_schema(functions, schema, memory))
        new_schema, new_memory = ourpipes.fast_forward_schema(schema, memory)
        assert executed["y"] == 4
    except Exception as e:
        assert False, str(e)    

def test_non_async_functions():

    def f(x, y):
        return x+y

    def g(x):
        return x*2

    memory = {
        "a": 1,
        "b": 2
    }
    schema = {
        "x": ("f", (("a", "x"), ("b", "y"))),
        "y": ("g", (("x", "x"),)),
    }
    functions = {
        "f": f,
        "g": g,
    }    

    try:
        executed = asyncio.run(ourpipes.execute_schema(functions, schema, memory, 'y'))
        assert executed == 6
    except Exception as e:
        assert False, str(e)  

def test_compiling_and_running_schema_with_kwargs():

    def f(x, y):
        return x+y

    def g(x, y):
        return x*y

    try:
        compiled = ourpipes.compile_schema_async(
            name="add_n_multiply", 
            description="Adding numbers followed by multiplication",
            functions={'f': f, 'g': g},
            schema={
                "x": ("f", (("a", "x"), ("b", "y"))),
                "y": ("g", (("x", "x"), ("c", "y"))),
            },
            executor=async_partial(
                ourpipes.execute_schema, 
                extract_key="y",
            ),
        )

        result = asyncio.run(compiled(a=3, b=2, c=2))
        assert result == 10

    except Exception as e:
        assert False, str(e)

def test_compiling_and_running_schema_with_args():

    def f(x, y):
        return x+y

    def g(x, y):
        return x*y

    try:
        compiled = ourpipes.compile_schema_async(
            name="add_n_multiply", 
            description="Adding numbers followed by multiplication",
            functions={'f': f, 'g': g},
            schema={
                "x": ("f", (("a", "x"), ("b", "y"))),
                "y": ("g", (("x", "x"), ("c", "y"))),
            },
            executor=async_partial(
                ourpipes.execute_schema, 
                extract_key="y",
            ),
        )

        result = asyncio.run(compiled(3, 2, 2))
        assert result == 10

    except Exception as e:
        assert False, str(e)

def test_automake_memory_from_schema():

    def f(x, y):
        return x+y

    def g(x, y):
        return x*y

    def h(x, y):
        return x-y

    try:
        functions = {k.__name__: k for k in [f, g, h]}
        schema = {
            'x': ('f', (('a', 'x'), ('b', 'y'))),
            'y': ('g', (('c', 'x'), ('d', 'y'))),
            'z': ('h', (('x', 'x'), ('y', 'y'))),
            'n': ('f', (('z', 'x'), ('z', 'y')))
        }
        memory = ourpipes.automake_memory(schema, args=[1,2,3,4])
        result = asyncio.run(ourpipes.execute_schema(functions, schema, memory, 'n'))
        assert result == -18
    except Exception as e:
        assert False, str(e)

def test_make_schema_from_string_and_run():

    def f(x, y):
        return x+y

    def g(x):
        return x*2

    def h(x, y):
        return x*y

    try:
        fnstr = "(a:x, b:y) -> f:x. (c:x) -> g:y. (x:x, y:y) -> h:z"
        schema = ourpipes.funstr2schema(fnstr)

        # Should be a, b, and c as arguments
        memory = {'a': 1, 'b': 2, 'c': 3}
        result = asyncio.run(ourpipes.execute_schema({k.__name__: k for k in [f, g, h]}, schema, memory, 'z'))

        assert result == 18
    except Exception as e:
        assert False, str(e)

def test_running_schema_with_args_kwargs_function_inputs():

    def f_args(*args):
        return sum(*args)

    def f_kwargs(**kwargs):
        return "".join(kwargs['b'].keys())

    def concat(thesum: int, thekeys: str) -> dict:
        return {thekeys: thesum}

    try:
        functions = ourpipes.fns2dict(f_args, f_kwargs, concat)
        schema = {
            'x': ("f_args", (('a'), )),
            'y': ("f_kwargs", (('b', 'b'), )),
            'z': ("concat", (('x', 'thesum'), ('y', 'thekeys'))),
        }
        memory = {
            'a': [1,2,3,4],
            'b': {'a': 1, 'b': 2, 'c': 3},
        }
        result = asyncio.run(ourpipes.execute_schema(functions, schema, memory, 'z'))
        assert result == {'abc': 10}
    except Exception as e:
        assert False, str(e)

def test_compile_multiple_functions_from_strings_and_run():

    def f(x, y, z):
        return x+y+z

    def g(x, y):
        return x-y

    def h(x, y):
        return x*y

    functions = ourpipes.fns2dict(f,g,h)
    schema_string_memories = [
        ("(a:x, b:y, c:z) -> f:x. (a:x, x:y) -> g:y. (b:x, y:y) -> h:z", {'a': 1, 'b': 2, 'c': 3}, 'z', -10),
        ("(a:x, b:y) -> h:x. (a:x, x:y) -> g:y. (b:x, y:y, x:z) -> f:z", {'a': 1, 'b': 2, 'c': 3}, 'z', 3),
        ("(a:x, b:y) -> h:x", {'a': 1, 'b': 2, 'c': 3}, 'x', 2),
    ]

    try:
        for funstr, memory, final, expected_result in schema_string_memories:
            actual_result = asyncio.run(
                ourpipes.execute_schema(functions, ourpipes.funstr2schema(funstr), memory, final),
            )
            assert expected_result == actual_result
    except Exception as e:
        assert False, str(e)

def test_compile_function_and_use_in_other_schema():

    def f(x, y):
        return x+y

    def g(x, y):
        return x*y

    functions = ourpipes.fns2dict(f, g)
    try:
        add_mul_add = ourpipes.compile_schema_async(
            name="add_mul_add", 
            functions=functions, 
            schema=ourpipes.funstr2schema("(a:x, b:y) -> f:x. (a:x, b:y) -> g:y. (x:x, y:y) -> f:z"), 
            executor=async_partial(
                ourpipes.execute_schema,
                extract_key="z",
            ),
        )

        functions[add_mul_add.__name__] = add_mul_add
        compiled = ourpipes.compile_schema_async(
            name="final_fn", 
            functions=functions, 
            schema=ourpipes.funstr2schema("(a:a, b:b) -> add_mul_add:x. (x:x, c:y) -> g:y"), 
            executor=async_partial(
                ourpipes.execute_schema,
                extract_key="y",
            ),
        )

        result = asyncio.run(compiled(a=1, b=2, c=3))
        assert result == 15
    except Exception as e:
        assert False, str(e)

def test_compile_function_use_in_other_schema_with_local_variables():

    def f(x, y):
        return x+y

    def g(x, y):
        return x*y

    functions = ourpipes.fns2dict(f, g)
    try:
        add_mul_add = ourpipes.compile_schema_async(
            name="add_mul_add", 
            functions=functions, 
            schema={
                "x": ["f", [["a", "x"], ["b", "y"]], {"b": 2}],
                "y": ["g", [["a", "x"], ["b", "y"]], {"b": 3}],
                "z": ["f", [["x", "x"], ["y", "y"]]]
            }, 
            executor=async_partial(
                ourpipes.execute_schema,
                extract_key="z",
            ),
        )

        functions[add_mul_add.__name__] = add_mul_add
        compiled = ourpipes.compile_schema_async(
            name="final_fn", 
            functions=functions, 
            schema={
                "x": ["add_mul_add", [["a", "a"]]], # no b here since it is in last schemas locals
                "y": ["g", [["x", "x"], ["c", "y"]]]
            }, 
            executor=async_partial(
                ourpipes.execute_schema,
                extract_key="y",
            ),
        )

        result = asyncio.run(compiled(a=1, c=3))
        assert result == 18
    except Exception as e:
        assert False, str(e)

def test_cut_schema():

    try:
        sub_schema = ourpipes.cut_schema(
            schema=ourpipes.funstr2schema("(a:x, b:y) -> f:x. (x:x) -> g:y. (x:x, y:y) -> h:z"),
            start=2,
        )
        assert len(sub_schema) == 2
        assert all((k in sub_schema for k in ["y", "z"]))
    except Exception as e:
        assert False, str(e)

def test_run_schema_with_insufficient_functions():

    def f(x):
        return x+1

    functions = ourpipes.fns2dict(f)
    schema = ourpipes.funstr2schema("(x:x) -> f:y. (y:y) -> g:z")
    try:
        # Should calculate as long as it can...
        result = asyncio.run(ourpipes.execute_schema(functions, schema, {"x": 2}))
        assert result['y'] == 3
        assert isinstance(result['z'], Exception)
    except Exception as e:
        assert False, str(e)

def test_funstr2schema_as_positional_arguments():

    def f(x):
        return x+1

    def g(x): 
        return x+2

    functions = ourpipes.fns2dict(f, g)
    try:
        schema = ourpipes.funstr2schema("(x) -> f:y. (y) -> g:z")
        result = asyncio.run(ourpipes.execute_schema(functions, schema, {"x": 2}))
        assert result['z'] == 5
    except Exception as e:
        assert False, str(e)

def test_local_constants_into_memory():

    try:
        schema = {
            "x": [
                "f",
                [
                    ["a", "x"],
                    ["b", "y"]
                ],
                {
                    "a": 1
                }
            ]
        }
        new_memory = ourpipes.local_constants2memory({}, schema)
        assert new_memory['x.f.a'] == 1
    except Exception as e:
        assert False, str(e)

def test_local_constants_into_memory_with_conflicting_variable_overriding():

    try:
        schema = {
            "x": [
                "f",
                [
                    ["a", "x"]
                ],
                {
                    "a": 1
                }
            ]
        }
        new_memory = ourpipes.local_constants2memory({'a': 2}, schema)
        assert new_memory['x.f.a'] == 1
    except Exception as e:
        assert False, str(e)

def test_local_constants_into_memory_with_conflicting_variable_not_overriding():

    try:
        schema = {
            "x": [
                "f",
                [
                    ["a", "x"]
                ],
                {
                    "a": 1
                }
            ]
        }
        new_memory = ourpipes.local_constants2memory({'x.f.a': 2}, schema, override=False)
        assert new_memory['x.f.a'] == 2
    except Exception as e:
        assert False, str(e)

def test_compute_schema_with_local_constants():

    def f(x):
        return x+1

    def g(x, y): 
        return x+y

    functions = ourpipes.fns2dict(f, g)
    schema = {
        "x": [
            "f",
            [
                ["a", "x"]
            ],
            {
                "a": 1
            }
        ],
        "y": [
            "g",
            [
                ["x", "x"],
                ["b", "y"]
            ],
            {
                "b": 2
            }
        ]
    }
    try:
        result = asyncio.run(ourpipes.execute_schema(functions, schema, {}))
        assert result['y'] == 4
    except Exception as e:
        assert False, str(e)

def test_schema2arguments_with_local_constants():

    def x(m):
        return m+1

    try:
        assert len(ourpipes.schema2arguments({
            "x": ["x", [["a"]], {"a": 1}],
            "y": ["x", [["x"]]]
        })) == 0

        assert len(ourpipes.schema2arguments({
            "x": ["x", [["a"]]],
            "y": ["x", [["x"]], {"x": 1}]
        })) == 1

        assert len(ourpipes.schema2arguments({
            "x": ["x", [["a"]]],
            "y": ["x", [["b"]]],
            "z": ["x", [["c"]]]
        })) == 3
    except Exception as e:
        assert False, str(e)

def test_function_does_not_exists():

    def f(x):
        return x+1

    schema = {
        "x": ["f", [["a"]]],
        "y": ["g", [["b"]]],
    }

    try:
        functions = ourpipes.fns2dict(f)
        results = asyncio.run(ourpipes.execute_schema(functions, schema, {"a": 1, "b": 2}))
        assert isinstance(results['y'], Exception)
    except Exception as e:
        assert False, str(e)

def test_argument_is_not_provided_when_asked():

    def f(x):
        return x+1

    schema = {
        "x": ["f", [["a"]]],
        "y": ["g", [["b"]]],
    }

    try:
        functions = ourpipes.fns2dict(f)
        asyncio.run(ourpipes.execute_schema(functions, schema, {}))
    except Exception as e:
        assert True, str(e)




import functools
import asyncio
import inspect
import itertools
import gzip
import pickle

from toposort import toposort
from typing import Callable, Tuple, Any

def schema2execution_order(schema: dict) -> iter:
    return toposort(
        {
            key: set(
                [
                    y[0] 
                    for y in fn[1]
                ]
            ) 
            for key, fn in schema.items()
        }
    )

def default_sequential_function_wrapper(fn_name: str, functions: dict, args, kwargs):
    """
        Let's check f and compute.

        Return:
            Any
    """
    async def wrap_fn(fn_name: str, functions: dict, *args, **kwargs):
        if not fn_name in functions:
            return Exception(f"function '{fn_name}' does not exists")

        fn = functions[fn_name]
        if inspect.iscoroutinefunction(fn):
            return await fn(*args, **kwargs)
        else:
            return fn(*args, **kwargs)

    return asyncio.create_task(wrap_fn(fn_name, functions, *args, **kwargs))

async def default_parallel_function_executor(*fns, **kwargs):

    """
        Uses asyncio to compute a list of functions.

        Return:
            List of function results
    """

    return await asyncio.gather(*fns, **kwargs)

async def asyncfn_executor_wrapper(fn, *args, **kwargs):

    """
        Checks if fn is a coroutine function, and if so
        awaits it. Else function is computed normally.

        Return:
            Function result
    """

    return await fn(*args, **kwargs) if inspect.iscoroutinefunction(fn) else fn(*args, **kwargs)

def select_function(fn_name: str, functions: dict) -> Callable:

    """
        Selecting function wrapper needed for when there's a
        key error in functions, just to append more info into
        the exception
    """
    if not fn_name in functions:
        return Exception(f"service doesn't offer function '{fn_name}'")

    return functions[fn_name]

def from_dicts(key: str, *dicts, default: Any = None):

    """
        Returns value of key in first matchning dict.
        If not matching dict, default value is returned.

        Return:
            Any
    """
    for d in dicts:
        if key in d:
            return d[key]

    return default

def from_lists(index: int, *lists, default: Any = {}):

    """
        Returns value of index in first matching list
        having index `index`. If no match, default
        is returned.

        Return:
            Any
    """
    for lst in lists:
        if len(lst) > index:
            return lst[index]

    return default

def local_constants2memory(memory: dict, schema: dict, override: bool = True) -> dict:

    """
        Checks for local constants in schema and transfer
        them into memory. If constant is already in memory, and `override`
        is set to true, value will be replaces. If false, it will not
        be replaced.

        Return:
            dict: memory
    """

    for var, fn_args in schema.items():
        fn_name, args = fn_args[0], fn_args[1]
        local_mem = from_lists(2, *[fn_args])

        if len(local_mem) > 0:
            for i, arg_key in enumerate(map(lambda x: x[0], args)):
                if arg_key in local_mem:
                    new_arg_key = ".".join([var, fn_name, arg_key])

                    # Only if override is True OR new_arg_key is not in memory already
                    if override or new_arg_key not in memory:
                        memory[new_arg_key] = local_mem[arg_key]

                    schema[var][1][i][0] = new_arg_key

    return memory



async def execute_schema(
    functions: dict,
    schema: dict, 
    memory: dict, 
    extract_key: str = None,

    paralell_function_executor: Callable = default_parallel_function_executor, 
    sequential_function_wrapper: Callable = default_sequential_function_wrapper,
) -> Tuple[str, dict]:

    """
        Executes functions in schema in topological sort order. If there is
        no such order, an exception is raised. Currently, the function does
        not type check so user needs to be careful when using. 

        Here's an example:

        functions:
            f : x, y -> x + y
            g : x, y -> x - y
            h : x, y -> x ^ y

        memory = {
            "a": 1,
            "b": 2,
            "c": 3
        }
        func_schema = {
            # output:, (func, ((memory_var, func_param_var), ...))
            "x": (f, (("a", "x"), ("b", "y"))),
            "y": (g, (("a", "x"), ("c", "y"))),
            "z": (h, (("x", "x"), ("y", "y"))),
        }

        Running this executor with given functions and inputs will result in {"z": 27}. 
        NOTE that the output variable name and function parameter variable must collaborate. 
        The result from a function will be saved in memory by the variable name of output 
        variable and may later be looked up by another function.

        We will return a tuple of (status, result), where the status is either "success" or
        "fail". When fail, the exception from underlying functions will be forward. Result
        is the latest function step computed.

        Return:
            Tuple[str, dict]
    """

    def read_function_from_variable(schema: dict, variable: str):

        try:
            return schema[variable][0]
        except KeyError:
            raise Exception(f"value for variable '{variable}' must be provided either in function result or in memory")
        except Exception as e:
            return e

    memory = local_constants2memory(memory, schema)
    execution_order = schema2execution_order(schema)
    
    for i, fn_pars in enumerate(execution_order):

        # Update memory
        memory.update(
            {
                variable: result
                for variable, result in zip(
                    filter(
                        lambda variable: not variable in memory,
                        fn_pars
                    ),
                    await paralell_function_executor(
                        *map(
                            lambda variable: sequential_function_wrapper(
                                fn_name=read_function_from_variable(schema, variable), 
                                functions=functions,
                                args=(
                                    memory[var[0]]
                                    for var in filter(
                                        lambda x: len(x) == 1,
                                        schema[variable][1]
                                    )
                                ),
                                kwargs={
                                    param: memory[var] 
                                    for var, param in filter(
                                        lambda x: len(x) == 2,
                                        schema[variable][1]
                                    )
                                },
                            ),
                            filter(
                                lambda variable: not variable in memory,
                                fn_pars
                            ),
                        ),
                        return_exceptions=True,
                    )
                )
            }
        )

    return memory[extract_key] if extract_key in memory else memory

def fast_forward_schema(
    schema: dict, 
    memory: dict, 
):
    """
        Will analyze scheme how far possible it can be forward, 
        which can later be executed. This can be necessary when 
        e.g this service doesn't offer all functions that scheme requires.
        
        Return:
            Tuple[memory, scheme]
    """   
    # Keep schema points if they either are not in memory
    # or if the memory spot is an Exception
    new_schema = {
        k: v 
        for k,v in schema.items() 
        if not k in memory or isinstance(memory[k], Exception)
    }

    new_memory = {
        k: v
        for k, v in memory.items()
        if any(
            itertools.chain(
                (
                    any(k == arg[0] for arg in args) 
                    for _, args in new_schema.values()
                ),
            ),
        )
    }

    return new_schema, new_memory


def default_decompression(compression_name: str, data):

    """
        Decompression mapping, maps compression string name
        to different decompression functions.

        Return:
            Any
    """

    return {
        'gzip':         gzip.decompress,
        'gzip_pickle':  lambda x: pickle.loads(gzip.decompress(x)),
    }.get(compression_name, 'gzip_pickle')(data)

def default_compression(compression_name: str, data):

    """
        Compression mapping, maps compression string name
        to different compression functions.

        Return:
            byte-string
    """
    return {
        'gzip':         gzip.compress,
        'gzip_pickle':  lambda x: gzip.compress(pickle.dumps(x, protocol=pickle.HIGHEST_PROTOCOL)),
    }.get(compression_name, 'gzip_pickle')(data)

async def execute_compressed_pipeline(
    functions: dict, 
    decompression_name: str, 
    payload: str, 
    decompress_map: Callable = default_decompression, 
    pipeline_executor: Callable = execute_schema,
):

    """
        execute_compressed_pipeline is a pipeline executor wrapper
        that expects a compressed pipeline in payload, defined how in header,
        and computes the pipeline.

        Return:
            Any
    """

    return await pipeline_executor(
        functions=functions,
        **decompress_map(
            decompression_name,
            payload, 
        ),
    )    

def compress_pipeline(compression_name: str, schema: dict, memory: dict, extract_key: str = None, compression_map: Callable = default_compression):

    """
        Compress pipeline into a bytes object.

        Return:
            byte-string
    """

    return compression_map(
        compression_name, 
        {
            'schema': schema, 
            'memory': memory,
            'extract_key': extract_key,
        },
    )

def schema2arguments(schema: dict) -> list:

    """
        Variables that are defined as function inputs,
        but are not declared as expecting from a function
        output, is called an argument.

        Return:
            list
    """

    return list({
        (i,j): inputs[0]
        for i, (output, fn_params) in enumerate(schema.items())
        for j, inputs in enumerate(fn_params[1])
        if not (inputs[0] in schema or inputs[0] in (fn_params[2] if len(fn_params) >= 3 else {}))
    }.values())

def schema2signature(schema: dict, functions: dict) -> tuple:

    """
        Returns some kind of artificial function
        signature from schema. The tuple consists
        of input argument names and return types.

        Return:
            Tuple[List[str], List[str]]
    """
    
    # Find all inputs that does not exists as a function
    # in schema
    arguments = schema2arguments(schema)
    execution_order = list(list(schema2execution_order(schema))[-1])
    return_types = [
        str(inspect.signature(functions[schema[output_name][0]]).return_annotation)
        for output_name in execution_order
    ]

    return (arguments, return_types)

def automake_memory(schema: dict, args: list = [], kwargs: dict = {}) -> dict:

    """
        Will generate schema memory input based on positional and keyword arguments.
        NOTE kwargs are in higher order than args and therefore preferred.

        Return:
            dict
    """
    arguments = schema2arguments(schema)
    for i, arg in enumerate(filter(lambda arg: arg not in kwargs, arguments)):
        if len(args) > i:
            kwargs[arg] = args[i]

    return kwargs

def compile_schema_async(
    name: str, 
    functions: dict, 
    schema: dict, 
    executor: Callable,
    description: str = "",
) -> Callable:

    """
        Compiling schema, functions and an executor into a virtual async function with name `name`.

        Return:
            Callable[dict]
    """

    def async_compiler(functions: dict, schema: dict, executor: Callable) -> Callable:

        async def wrap_executor(*args, **kwargs) -> Any:
            return await executor(
                functions=functions, 
                schema=schema, 
                memory=automake_memory(
                    schema=schema, 
                    args=args, 
                    kwargs=kwargs,
                ),
            )
        
        wrap_executor.__name__ = name
        wrap_executor.__qualname__ = name
        wrap_executor.__doc__ = description
        return wrap_executor

    return async_compiler(functions, schema, executor)

def funstr2schema(funstr: str) -> dict:

    """
        A `funstr` is a functional-string, containing information
        of a schema in one string, written in a certain way.

        Explinations: 
            Let funstr = (a:x, b:y) -> f:x. (c:x) -> g. (f:x, g:y) -> h:y 
            (:x, :y) -> f mening that f is expected to have x, y as keyword
            input. (a:x, b:y) mening variable a maps to x, and variable b to y.
            (:, :) -> f meaning that f is expected to have two positional arg inputs.
            f:x meaning that result from ... -> f will be stored in variable x.
            Dot (.) meaning end of statement.
    """

    try:
        schema = {}
        statements = funstr.replace(" ", "").split(".")
        for statement in statements:
            arguments_string, function_string = statement.split("->")
            poskey_arguments = arguments_string.replace("(", "").replace(")", "").split(",")
            arguments = [tuple(arg.split(':')) for arg in poskey_arguments]
            fn_name, outvar = function_string.split(":")

            schema[outvar] = (fn_name, arguments)
        
        return schema
    except Exception as e:
        raise Exception(f"could not create schema from function-string '{funstr}' because of error: {e}")

def fns2dict(*functions) -> dict:

    """
        Returns a dictionary of function name -> function,
        given functions as *arguments.

        Return:
            Dict[str, Callable]
    """

    return {f.__name__: f for f in functions}

def cut_schema(schema: dict, start: int = 0, stop: int = None) -> dict:

    """
        Cut schema from point `start` to point `stop`.
        Starts default is 0, and stops default is at end.

        Return:
            dict (schema)
    """
    
    schema_order = list(schema2execution_order(schema))
    stop = stop if stop is not None else len(schema_order)

    schema_order_cut = schema_order[start:stop]
    schema_cut = {
        k: v
        for k, v in schema.items()
        if any((k in fns for fns in schema_order_cut))
    }
    return schema_cut



import asyncio
import functools
from typing import Callable

def async_partial(fn: Callable, *args, **kwargs):

    """
        Wraps functools `partial` into a async function.

        Return:
            coro Callable
    """
    partial_fn = functools.partial(fn, *args, **kwargs)
    async def runner(*inner_args, **inner_kwargs):
        return await partial_fn(*inner_args, **inner_kwargs)

    return runner